/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 David N. Welton.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        David N. Welton (http://www.dedasys.com/davidw/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "TclMM" and "David N. Welton" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    davidw@apache.org.
 *
 * 5. Products derived from this software may not be called "TclMM",
 *    nor may "TclMM" appear in their name, without prior written
 *    permission of the author(s).
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * ==================================================================== */

/*
 *-----------------------------------------------------------------------------
 *
 * TclMM -- Tcl interface to 'libmm' shared memory access library.
 *
 * High-level interface to 'libmm', using a hash table to provide
 * access to variables stored in shared memory.
 *
 *-----------------------------------------------------------------------------
 */

#include <tcl.h>
#include "tclHash.h"
#include <mm.h>

static Tcl_HashTable *tablePtr;

int TclMM_Init(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
    int sz = 0;

    if (objc != 3)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "size file");
	return TCL_ERROR;
    }

    if (Tcl_GetIntFromObj(interp, objv[1], &sz) != TCL_OK) {
	return TCL_ERROR;
    }
    MM_create(sz, Tcl_GetString(objv[2]));
/*     MM_permission(0600, getuid(), getgid());  */

    tablePtr = MM_malloc(sizeof(Tcl_HashTable));

    if (tablePtr == NULL)
    {
	Tcl_AddErrorInfo(interp, MM_error());
	return TCL_ERROR;
    }
    TclMM_InitHashTable(tablePtr, TCL_STRING_KEYS);

    return TCL_OK;
}

int TclMM_End(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1)
    {
	Tcl_WrongNumArgs(interp, 1, objv, NULL);
	return TCL_ERROR;
    }
    TclMM_DeleteHashTable(tablePtr);
    MM_free(tablePtr);
    MM_destroy();
    return TCL_OK;
}

int TclMM_Set(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
    if (objc == 2 || objc == 3)
    {
	char *varname;
	char *val;
	int newPtr = 0;
	Tcl_HashEntry *entryPtr;

	MM_lock(MM_LOCK_RW);
	varname = Tcl_GetString(objv[1]);
	entryPtr = TclMM_CreateHashEntry(tablePtr, varname, &newPtr);

	if (objc == 3)
	{
	    val = Tcl_GetString(objv[2]);
	    TclMM_SetHashValue(entryPtr, val);
	    Tcl_SetResult(interp, val, NULL);
	} else {
	    if (newPtr)
	    {
		TclMM_DeleteHashEntry(entryPtr);
		Tcl_AddErrorInfo(interp, "variable doesn't exist");
		MM_unlock();
		return TCL_ERROR;
	    }
	    val = TclMM_GetHashValue(entryPtr);
	    Tcl_SetResult(interp, val, NULL);
	}
	MM_unlock();
    }
    else
    {
	Tcl_WrongNumArgs(interp, 1, objv, "varname ?value?");
	return TCL_ERROR;
    }
    return TCL_OK;
}

int TclMM_UnSet(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
    int newPtr = 0;
    Tcl_HashEntry *entryPtr;

    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "varname");
    }

    MM_lock(MM_LOCK_RW);
    entryPtr = TclMM_CreateHashEntry(tablePtr, Tcl_GetString(objv[1]), &newPtr);
    if (newPtr) {
	Tcl_AddErrorInfo(interp, "no such variable");
	TclMM_DeleteHashEntry(entryPtr);
	MM_unlock();
	return TCL_ERROR;
    }
    TclMM_DeleteHashEntry(entryPtr);

    MM_unlock();
    return TCL_OK;
}

int TclMM_Exists(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
    int newPtr = 0;
    Tcl_HashEntry *entryPtr;

    if (objc != 2)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "varname");
    }

    MM_lock(MM_LOCK_RW);
    entryPtr = TclMM_CreateHashEntry(tablePtr, Tcl_GetString(objv[1]), &newPtr);
    if (newPtr) {
	Tcl_SetObjResult(interp, Tcl_NewIntObj(0));
    } else {
	Tcl_SetObjResult(interp, Tcl_NewIntObj(1));
    }
    TclMM_DeleteHashEntry(entryPtr);
    MM_unlock();
    return TCL_OK;
}

int TclMM_Available(ClientData clientData, Tcl_Interp *interp,
		    int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "");
	return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, Tcl_NewIntObj(MM_available()));
    return TCL_OK;
}

int TclMM_MaxSize(ClientData clientData, Tcl_Interp *interp,
		    int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1)
    {
	Tcl_WrongNumArgs(interp, 1, objv, "");
	return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, Tcl_NewLongObj(MM_maxsize()));
    return TCL_OK;
}

Tclmm_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "mm_init", TclMM_Init,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_end", TclMM_End,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_set", TclMM_Set,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_unset", TclMM_UnSet,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_exists", TclMM_Exists,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_available", TclMM_Available,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateObjCommand(interp, "mm_maxsize", TclMM_MaxSize,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_PkgProvide(interp, "TclMM", "0.1");

    return TCL_OK;
}
